@@ -8,6 +8,7 @@ use std::time::SystemTime;
 use fs_err as fs;
 use fs_err::{DirEntry, File};
 use reflink_copy as reflink;
+use serde::{Deserialize, Serialize};
 use tempfile::tempdir_in;
 use tracing::{debug, instrument};
 
@@ -201,7 +202,8 @@ fn parse_scripts(
     scripts_from_ini(extras, python_minor, ini)
 }
 
-#[derive(Debug, Clone, Copy)]
+#[derive(Debug, Clone, Copy, Serialize, Deserialize)]
+#[serde(deny_unknown_fields, rename_all = "kebab-case")]
 #[cfg_attr(feature = "clap", derive(clap::ValueEnum))]
 pub enum LinkMode {
     /// Clone (i.e., copy-on-write) packages from the wheel into the site packages.
@@ -333,14 +335,17 @@ fn clone_recursive(
                         if reflink::reflink(&from, &tempfile).is_ok() {
                             fs::rename(&tempfile, to)?;
                         } else {
-                            debug!("Failed to clone {} to temporary location {} - attempting to copy files as a fallback", from.display(), tempfile.display());
+                            debug!(
+                                "Failed to clone `{}` to temporary location `{}`, attempting to copy files as a fallback",
+                                from.display(),
+                                tempfile.display());
                             *attempt = Attempt::UseCopyFallback;
                             fs::copy(&from, &to)?;
                         }
                     }
                 } else {
                     debug!(
-                        "Failed to clone {} to {} - attempting to copy files as a fallback",
+                        "Failed to clone `{}` to `{}`, attempting to copy files as a fallback",
                         from.display(),
                         to.display()
                     );
@@ -463,10 +468,20 @@ fn hardlink_wheel_files(
                         if fs::hard_link(path, &tempfile).is_ok() {
                             fs_err::rename(&tempfile, &out_path)?;
                         } else {
+                            debug!(
+                                "Failed to hardlink `{}` to `{}`, attempting to copy files as a fallback",
+                                out_path.display(),
+                                path.display()
+                            );
                             fs::copy(path, &out_path)?;
                             attempt = Attempt::UseCopyFallback;
                         }
                     } else {
+                        debug!(
+                            "Failed to hardlink `{}` to `{}`, attempting to copy files as a fallback",
+                            out_path.display(),
+                            path.display()
+                        );
                         fs::copy(path, &out_path)?;
                         attempt = Attempt::UseCopyFallback;
                     }
