@@ -1,11 +1,11 @@
 use std::collections::hash_map::Entry;
 use std::hash::BuildHasherDefault;
-use std::io;
 
 use anyhow::{bail, Result};
 use rustc_hash::FxHashMap;
 use tracing::{debug, warn};
 
+use distribution_types::Hashed;
 use distribution_types::{
     BuiltDist, CachedDirectUrlDist, CachedDist, Dist, IndexLocations, InstalledDist,
     InstalledMetadata, InstalledVersion, Name, SourceDist,
@@ -14,9 +14,12 @@ use pep508_rs::{Requirement, VersionOrUrl};
 use platform_tags::Tags;
 use uv_cache::{ArchiveTarget, ArchiveTimestamp, Cache, CacheBucket, WheelCache};
 use uv_configuration::{NoBinary, Reinstall};
-use uv_distribution::{BuiltWheelIndex, RegistryWheelIndex};
+use uv_distribution::{
+    BuiltWheelIndex, HttpArchivePointer, LocalArchivePointer, RegistryWheelIndex,
+};
 use uv_fs::Simplified;
 use uv_interpreter::PythonEnvironment;
+use uv_types::HashStrategy;
 
 use crate::{ResolvedEditable, SitePackages};
 
@@ -53,20 +56,25 @@ impl<'a> Planner<'a> {
     /// plan will respect cache entries created after the current time (as per the [`Refresh`]
     /// policy). Otherwise, entries will be ignored. The downstream distribution database may still
     /// read those entries from the cache after revalidating them.
+    ///
+    /// The install plan will also respect the required hashes, such that it will never return a
+    /// cached distribution that does not match the required hash. Like pip, though, it _will_
+    /// return an _installed_ distribution that does not match the required hash.
     #[allow(clippy::too_many_arguments)]
     pub fn build(
         self,
         mut site_packages: SitePackages<'_>,
         reinstall: &Reinstall,
         no_binary: &NoBinary,
+        hasher: &HashStrategy,
         index_locations: &IndexLocations,
         cache: &Cache,
         venv: &PythonEnvironment,
         tags: &Tags,
     ) -> Result<Plan> {
         // Index all the already-downloaded wheels in the cache.
-        let mut registry_index = RegistryWheelIndex::new(cache, tags, index_locations);
-        let built_index = BuiltWheelIndex::new(cache, tags);
+        let mut registry_index = RegistryWheelIndex::new(cache, tags, index_locations, hasher);
+        let built_index = BuiltWheelIndex::new(cache, tags, hasher);
 
         let mut cached = vec![];
         let mut remote = vec![];
@@ -206,16 +214,9 @@ impl<'a> Planner<'a> {
                     }
                 }
                 Some(VersionOrUrl::VersionSpecifier(specifier)) => {
-                    if let Some(distribution) =
-                        registry_index
-                            .get(&requirement.name)
-                            .find_map(|(version, distribution)| {
-                                if specifier.contains(version) {
-                                    Some(distribution)
-                                } else {
-                                    None
-                                }
-                            })
+                    if let Some((_version, distribution)) = registry_index
+                        .get(&requirement.name)
+                        .find(|(version, _)| specifier.contains(version))
                     {
                         debug!("Requirement already cached: {distribution}");
                         cached.push(CachedDist::Registry(distribution.clone()));
@@ -252,24 +253,23 @@ impl<'a> Planner<'a> {
                                     CacheBucket::Wheels,
                                     WheelCache::Url(&wheel.url).wheel_dir(wheel.name().as_ref()),
                                 )
-                                .entry(wheel.filename.stem());
+                                .entry(format!("{}.http", wheel.filename.stem()));
 
-                            match cache_entry.path().canonicalize() {
-                                Ok(archive) => {
+                            // Read the HTTP pointer.
+                            if let Some(pointer) = HttpArchivePointer::read_from(&cache_entry)? {
+                                let archive = pointer.into_archive();
+                                if archive.satisfies(hasher.get(&wheel)) {
                                     let cached_dist = CachedDirectUrlDist::from_url(
                                         wheel.filename,
                                         wheel.url,
-                                        archive,
+                                        archive.hashes,
+                                        cache.archive(&archive.id),
                                     );
 
                                     debug!("URL wheel requirement already cached: {cached_dist}");
                                     cached.push(CachedDist::Url(cached_dist));
                                     continue;
                                 }
-                                Err(err) if err.kind() == io::ErrorKind::NotFound => {
-                                    // The cache entry doesn't exist, so it's not fresh.
-                                }
-                                Err(err) => return Err(err.into()),
                             }
                         }
                         Dist::Built(BuiltDist::Path(wheel)) => {
@@ -294,31 +294,27 @@ impl<'a> Planner<'a> {
                                     CacheBucket::Wheels,
                                     WheelCache::Url(&wheel.url).wheel_dir(wheel.name().as_ref()),
                                 )
-                                .entry(wheel.filename.stem());
+                                .entry(format!("{}.rev", wheel.filename.stem()));
 
-                            match cache_entry.path().canonicalize() {
-                                Ok(archive) => {
-                                    if ArchiveTimestamp::up_to_date_with(
-                                        &wheel.path,
-                                        ArchiveTarget::Cache(&archive),
-                                    )? {
+                            if let Some(pointer) = LocalArchivePointer::read_from(&cache_entry)? {
+                                let timestamp = ArchiveTimestamp::from_file(&wheel.path)?;
+                                if pointer.is_up_to_date(timestamp) {
+                                    let archive = pointer.into_archive();
+                                    if archive.satisfies(hasher.get(&wheel)) {
                                         let cached_dist = CachedDirectUrlDist::from_url(
                                             wheel.filename,
                                             wheel.url,
-                                            archive,
+                                            archive.hashes,
+                                            cache.archive(&archive.id),
                                         );
 
                                         debug!(
-                                            "URL wheel requirement already cached: {cached_dist}"
+                                            "Path wheel requirement already cached: {cached_dist}"
                                         );
                                         cached.push(CachedDist::Url(cached_dist));
                                         continue;
                                     }
                                 }
-                                Err(err) if err.kind() == io::ErrorKind::NotFound => {
-                                    // The cache entry doesn't exist, so it's not fresh.
-                                }
-                                Err(err) => return Err(err.into()),
                             }
                         }
                         Dist::Source(SourceDist::DirectUrl(sdist)) => {
