@@ -6,11 +6,11 @@ use std::path::PathBuf;
 use tracing::{debug, instrument};
 
 use uv_cache::Cache;
-use uv_fs::normalize_path;
+use uv_toolchain::PythonVersion;
 
 use crate::interpreter::InterpreterInfoError;
 use crate::python_environment::{detect_python_executable, detect_virtual_env};
-use crate::{Error, Interpreter, PythonVersion};
+use crate::{Error, Interpreter};
 
 /// Find a Python of a specific version, a binary with a name or a path to a binary.
 ///
@@ -43,17 +43,35 @@ pub fn find_requested_python(request: &str, cache: &Cache) -> Result<Option<Inte
             _ => unreachable!(),
         };
         find_python(selector, cache)
-    } else if !request.contains(std::path::MAIN_SEPARATOR) {
-        // `-p python3.10`; Generally not used on windows because all Python are `python.exe`.
-        let Some(executable) = find_executable(request)? else {
-            return Ok(None);
-        };
-        Interpreter::query(executable, cache).map(Some)
     } else {
-        // `-p /home/ferris/.local/bin/python3.10`
-        let executable = normalize_path(request);
-
-        Interpreter::query(executable, cache).map(Some)
+        match fs_err::metadata(request) {
+            Ok(metadata) => {
+                // Map from user-provided path to an executable.
+                let path = uv_fs::absolutize_path(request.as_ref())?;
+                let executable = if metadata.is_dir() {
+                    // If the user provided a directory, assume it's a virtual environment.
+                    // `-p /home/ferris/.venv`
+                    if cfg!(windows) {
+                        Cow::Owned(path.join("Scripts/python.exe"))
+                    } else {
+                        Cow::Owned(path.join("bin/python"))
+                    }
+                } else {
+                    // Otherwise, assume it's a Python executable.
+                    // `-p /home/ferris/.local/bin/python3.10`
+                    path
+                };
+                Interpreter::query(executable, cache).map(Some)
+            }
+            Err(err) if err.kind() == std::io::ErrorKind::NotFound => {
+                // `-p python3.10`; Generally not used on windows because all Python are `python.exe`.
+                let Some(executable) = find_executable(request)? else {
+                    return Ok(None);
+                };
+                Interpreter::query(executable, cache).map(Some)
+            }
+            Err(err) => return Err(err.into()),
+        }
     }
 }
 
@@ -464,7 +482,7 @@ fn find_version(
     let version_matches = |interpreter: &Interpreter| -> bool {
         if let Some(python_version) = python_version {
             // If a patch version was provided, check for an exact match
-            python_version.is_satisfied_by(interpreter)
+            interpreter.satisfies(python_version)
         } else {
             // The version always matches if one was not provided
             true
@@ -714,19 +732,15 @@ mod windows {
         #[cfg_attr(not(windows), ignore)]
         fn no_such_python_path() {
             let result =
-                find_requested_python(r"C:\does\not\exists\python3.12", &Cache::temp().unwrap());
-            insta::with_settings!({
-                filters => vec![
-                    // The exact message is host language dependent
-                    (r"Caused by: .* \(os error 3\)", "Caused by: The system cannot find the path specified. (os error 3)")
-                ]
-            }, {
-                assert_snapshot!(
-                    format_err(result), @r###"
-        failed to canonicalize path `C:\does\not\exists\python3.12`
-          Caused by: The system cannot find the path specified. (os error 3)
-        "###);
-            });
+                find_requested_python(r"C:\does\not\exists\python3.12", &Cache::temp().unwrap())
+                    .unwrap()
+                    .ok_or(Error::RequestedPythonNotFound(
+                        r"C:\does\not\exists\python3.12".to_string(),
+                    ));
+            assert_snapshot!(
+                format_err(result),
+                @"Failed to locate Python interpreter at `C:\\does\\not\\exists\\python3.12`"
+            );
         }
     }
 }
@@ -776,11 +790,12 @@ mod tests {
     #[test]
     #[cfg_attr(not(unix), ignore)]
     fn no_such_python_path() {
-        let result = find_requested_python("/does/not/exists/python3.12", &Cache::temp().unwrap());
+        let result = find_requested_python("/does/not/exists/python3.12", &Cache::temp().unwrap())
+            .unwrap()
+            .ok_or(Error::RequestedPythonNotFound(
+                "/does/not/exists/python3.12".to_string(),
+            ));
         assert_snapshot!(
-            format_err(result), @r###"
-        failed to canonicalize path `/does/not/exists/python3.12`
-          Caused by: No such file or directory (os error 2)
-        "###);
+            format_err(result), @"Failed to locate Python interpreter at `/does/not/exists/python3.12`");
     }
 }
