@@ -12,7 +12,7 @@ use pubgrub::type_aliases::SelectedDependencies;
 use rustc_hash::{FxHashMap, FxHashSet};
 
 use distribution_types::{
-    Dist, DistributionMetadata, LocalEditable, Name, PackageId, ResolvedDist, Verbatim,
+    Dist, DistributionMetadata, IndexUrl, LocalEditable, Name, ResolvedDist, Verbatim, VersionId,
     VersionOrUrl,
 };
 use once_map::OnceMap;
@@ -35,6 +35,11 @@ use crate::{Manifest, ResolveError};
 /// package.
 #[derive(Debug, Default, Copy, Clone, PartialEq)]
 #[cfg_attr(feature = "clap", derive(clap::ValueEnum))]
+#[cfg_attr(feature = "serde", derive(serde::Deserialize))]
+#[cfg_attr(
+    feature = "serde",
+    serde(deny_unknown_fields, rename_all = "kebab-case")
+)]
 pub enum AnnotationStyle {
     /// Render the annotations on a single, comma-separated line.
     Line,
@@ -66,7 +71,7 @@ impl ResolutionGraph {
         selection: &SelectedDependencies<UvDependencyProvider>,
         pins: &FilePins,
         packages: &OnceMap<PackageName, VersionsResponse>,
-        distributions: &OnceMap<PackageId, MetadataResponse>,
+        distributions: &OnceMap<VersionId, MetadataResponse>,
         state: &State<UvDependencyProvider>,
         preferences: &Preferences,
         editables: Editables,
@@ -96,14 +101,17 @@ impl ResolutionGraph {
 
                     // Add its hashes to the index, preserving those that were already present in
                     // the lockfile if necessary.
-                    if let Some(hash) = preferences.match_hashes(package_name, version) {
-                        hashes.insert(package_name.clone(), hash.to_vec());
+                    if let Some(digests) = preferences
+                        .match_hashes(package_name, version)
+                        .filter(|digests| !digests.is_empty())
+                    {
+                        hashes.insert(package_name.clone(), digests.to_vec());
                     } else if let Some(versions_response) = packages.get(package_name) {
                         if let VersionsResponse::Found(ref version_maps) = *versions_response {
                             for version_map in version_maps {
-                                if let Some(mut hash) = version_map.hashes(version) {
-                                    hash.sort_unstable();
-                                    hashes.insert(package_name.clone(), hash);
+                                if let Some(mut digests) = version_map.hashes(version) {
+                                    digests.sort_unstable();
+                                    hashes.insert(package_name.clone(), digests);
                                     break;
                                 }
                             }
@@ -126,17 +134,18 @@ impl ResolutionGraph {
 
                     // Add its hashes to the index, preserving those that were already present in
                     // the lockfile if necessary.
-                    if let Some(hash) = preferences.match_hashes(package_name, version) {
-                        hashes.insert(package_name.clone(), hash.to_vec());
-                    } else if let Some(versions_response) = packages.get(package_name) {
-                        if let VersionsResponse::Found(ref version_maps) = *versions_response {
-                            for version_map in version_maps {
-                                if let Some(mut hash) = version_map.hashes(version) {
-                                    hash.sort_unstable();
-                                    hashes.insert(package_name.clone(), hash);
-                                    break;
-                                }
-                            }
+                    if let Some(digests) = preferences
+                        .match_hashes(package_name, version)
+                        .filter(|digests| !digests.is_empty())
+                    {
+                        hashes.insert(package_name.clone(), digests.to_vec());
+                    } else if let Some(metadata_response) =
+                        distributions.get(&pinned_package.version_id())
+                    {
+                        if let MetadataResponse::Found(ref archive) = *metadata_response {
+                            let mut digests = archive.hashes.clone();
+                            digests.sort_unstable();
+                            hashes.insert(package_name.clone(), digests);
                         }
                     }
 
@@ -164,21 +173,21 @@ impl ResolutionGraph {
                             });
                         }
                     } else {
-                        let response = distributions.get(&dist.package_id()).unwrap_or_else(|| {
+                        let response = distributions.get(&dist.version_id()).unwrap_or_else(|| {
                             panic!(
                                 "Every package should have metadata: {:?}",
-                                dist.package_id()
+                                dist.version_id()
                             )
                         });
 
-                        let MetadataResponse::Found(metadata) = &*response else {
+                        let MetadataResponse::Found(archive) = &*response else {
                             panic!(
                                 "Every package should have metadata: {:?}",
-                                dist.package_id()
+                                dist.version_id()
                             )
                         };
 
-                        if metadata.provides_extras.contains(extra) {
+                        if archive.metadata.provides_extras.contains(extra) {
                             extras
                                 .entry(package_name.clone())
                                 .or_insert_with(Vec::new)
@@ -218,21 +227,21 @@ impl ResolutionGraph {
                             });
                         }
                     } else {
-                        let response = distributions.get(&dist.package_id()).unwrap_or_else(|| {
+                        let response = distributions.get(&dist.version_id()).unwrap_or_else(|| {
                             panic!(
                                 "Every package should have metadata: {:?}",
-                                dist.package_id()
+                                dist.version_id()
                             )
                         });
 
-                        let MetadataResponse::Found(metadata) = &*response else {
+                        let MetadataResponse::Found(archive) = &*response else {
                             panic!(
                                 "Every package should have metadata: {:?}",
-                                dist.package_id()
+                                dist.version_id()
                             )
                         };
 
-                        if metadata.provides_extras.contains(extra) {
+                        if archive.metadata.provides_extras.contains(extra) {
                             extras
                                 .entry(package_name.clone())
                                 .or_insert_with(Vec::new)
@@ -425,23 +434,23 @@ impl ResolutionGraph {
         let mut seen_marker_values = FxHashSet::default();
         for i in self.petgraph.node_indices() {
             let dist = &self.petgraph[i];
-            let package_id = match dist.version_or_url() {
+            let version_id = match dist.version_or_url() {
                 VersionOrUrl::Version(version) => {
-                    PackageId::from_registry(dist.name().clone(), version.clone())
+                    VersionId::from_registry(dist.name().clone(), version.clone())
                 }
-                VersionOrUrl::Url(verbatim_url) => PackageId::from_url(verbatim_url.raw()),
+                VersionOrUrl::Url(verbatim_url) => VersionId::from_url(verbatim_url.raw()),
             };
             let res = index
                 .distributions
-                .get(&package_id)
+                .get(&version_id)
                 .expect("every package in resolution graph has metadata");
-            let MetadataResponse::Found(md) = &*res else {
+            let MetadataResponse::Found(archive, ..) = &*res else {
                 panic!(
                     "Every package should have metadata: {:?}",
-                    dist.package_id()
+                    dist.version_id()
                 )
             };
-            for req in manifest.apply(&md.requires_dist) {
+            for req in manifest.apply(&archive.metadata.requires_dist) {
                 let Some(ref marker_tree) = req.marker else {
                     continue;
                 };
@@ -493,6 +502,7 @@ impl ResolutionGraph {
 
 /// A [`std::fmt::Display`] implementation for the resolution graph.
 #[derive(Debug)]
+#[allow(clippy::struct_excessive_bools)]
 pub struct DisplayResolutionGraph<'a> {
     /// The underlying graph.
     resolution: &'a ResolutionGraph,
@@ -505,6 +515,8 @@ pub struct DisplayResolutionGraph<'a> {
     /// Whether to include annotations in the output, to indicate which dependency or dependencies
     /// requested each package.
     include_annotations: bool,
+    /// Whether to include indexes in the output, to indicate which index was used for each package.
+    include_index_annotation: bool,
     /// The style of annotation comments, used to indicate the dependencies that requested each
     /// package.
     annotation_style: AnnotationStyle,
@@ -518,6 +530,7 @@ impl<'a> From<&'a ResolutionGraph> for DisplayResolutionGraph<'a> {
             false,
             false,
             true,
+            false,
             AnnotationStyle::default(),
         )
     }
@@ -525,12 +538,14 @@ impl<'a> From<&'a ResolutionGraph> for DisplayResolutionGraph<'a> {
 
 impl<'a> DisplayResolutionGraph<'a> {
     /// Create a new [`DisplayResolutionGraph`] for the given graph.
+    #[allow(clippy::fn_params_excessive_bools)]
     pub fn new(
         underlying: &'a ResolutionGraph,
         no_emit_packages: &'a [PackageName],
         show_hashes: bool,
         include_extras: bool,
         include_annotations: bool,
+        include_index_annotation: bool,
         annotation_style: AnnotationStyle,
     ) -> DisplayResolutionGraph<'a> {
         Self {
@@ -539,6 +554,7 @@ impl<'a> DisplayResolutionGraph<'a> {
             show_hashes,
             include_extras,
             include_annotations,
+            include_index_annotation,
             annotation_style,
         }
     }
@@ -576,6 +592,14 @@ impl<'a> Node<'a> {
             Node::Distribution(name, _, _) => NodeKey::Distribution(name),
         }
     }
+
+    /// Return the [`IndexUrl`] of the distribution, if any.
+    fn index(&self) -> Option<&IndexUrl> {
+        match self {
+            Node::Editable(_, _) => None,
+            Node::Distribution(_, dist, _) => dist.index(),
+        }
+    }
 }
 
 impl Verbatim for Node<'_> {
@@ -660,6 +684,8 @@ impl std::fmt::Display for DisplayResolutionGraph<'_> {
             // Determine the annotation comment and separator (between comment and requirement).
             let mut annotation = None;
 
+            // If enabled, include annotations to indicate the dependencies that requested each
+            // package (e.g., `# via mypy`).
             if self.include_annotations {
                 // Display all dependencies.
                 let mut edges = self
@@ -714,6 +740,14 @@ impl std::fmt::Display for DisplayResolutionGraph<'_> {
                 // Write the line as is.
                 writeln!(f, "{line}")?;
             }
+
+            // If enabled, include indexes to indicate which index was used for each package (e.g.,
+            // `# from https://pypi.org/simple`).
+            if self.include_index_annotation {
+                if let Some(index) = node.index() {
+                    writeln!(f, "{}", format!("    # from {index}").green())?;
+                }
+            }
         }
 
         Ok(())
