@@ -14,11 +14,11 @@ use pubgrub::range::Range;
 use pubgrub::solver::{Incompatibility, State};
 use rustc_hash::{FxHashMap, FxHashSet};
 use tokio_stream::wrappers::ReceiverStream;
-use tracing::{debug, info_span, instrument, trace, warn, Instrument};
+use tracing::{debug, enabled, info_span, instrument, trace, warn, Instrument, Level};
 
 use distribution_types::{
     BuiltDist, Dist, DistributionMetadata, IncompatibleDist, IncompatibleSource, IncompatibleWheel,
-    InstalledDist, Name, RemoteSource, ResolvedDist, ResolvedDistRef, SourceDist, VersionOrUrl,
+    InstalledDist, RemoteSource, ResolvedDist, ResolvedDistRef, SourceDist, VersionOrUrl,
 };
 pub(crate) use locals::Locals;
 use pep440_rs::{Version, MIN_VERSION};
@@ -26,12 +26,12 @@ use pep508_rs::{MarkerEnvironment, Requirement};
 use platform_tags::Tags;
 use pypi_types::Metadata23;
 pub(crate) use urls::Urls;
-use uv_client::{FlatIndex, RegistryClient};
+use uv_client::RegistryClient;
 use uv_configuration::{Constraints, Overrides};
-use uv_distribution::DistributionDatabase;
+use uv_distribution::{ArchiveMetadata, DistributionDatabase};
 use uv_interpreter::Interpreter;
 use uv_normalize::PackageName;
-use uv_types::{BuildContext, InstalledPackagesProvider};
+use uv_types::{BuildContext, HashStrategy, InstalledPackagesProvider};
 
 use crate::candidate_selector::{CandidateDist, CandidateSelector};
 use crate::editables::Editables;
@@ -54,7 +54,7 @@ pub use crate::resolver::provider::{
 use crate::resolver::reporter::Facade;
 pub use crate::resolver::reporter::{BuildId, Reporter};
 use crate::yanks::AllowedYanks;
-use crate::{DependencyMode, Exclusions, Options};
+use crate::{DependencyMode, Exclusions, FlatIndex, Options};
 
 mod batch_prefetch;
 mod index;
@@ -71,19 +71,32 @@ pub(crate) enum UnavailableVersion {
     IncompatibleDist(IncompatibleDist),
 }
 
-/// The package is unavailable and cannot be used
+/// The package is unavailable and cannot be used.
 #[derive(Debug, Clone)]
 pub(crate) enum UnavailablePackage {
-    /// Index lookups were disabled (i.e., `--no-index`) and the package was not found in a flat index (i.e. from `--find-links`)
+    /// Index lookups were disabled (i.e., `--no-index`) and the package was not found in a flat index (i.e. from `--find-links`).
     NoIndex,
     /// Network requests were disabled (i.e., `--offline`), and the package was not found in the cache.
     Offline,
-    /// The package was not found in the registry
+    /// The package was not found in the registry.
     NotFound,
+    /// The package metadata was found, but could not be parsed.
+    InvalidMetadata(String),
+    /// The package has an invalid structure.
+    InvalidStructure(String),
+}
+
+/// The package is unavailable at specific versions.
+#[derive(Debug, Clone)]
+pub(crate) enum IncompletePackage {
+    /// Network requests were disabled (i.e., `--offline`), and the wheel metadata was not found in the cache.
+    Offline,
     /// The wheel metadata was found, but could not be parsed.
-    InvalidMetadata,
+    InvalidMetadata(String),
+    /// The wheel metadata was found, but the metadata was inconsistent.
+    InconsistentMetadata(String),
     /// The wheel has an invalid structure.
-    InvalidStructure,
+    InvalidStructure(String),
 }
 
 enum ResolverVersion {
@@ -108,13 +121,16 @@ pub struct Resolver<
     urls: Urls,
     locals: Locals,
     dependency_mode: DependencyMode,
+    hasher: &'a HashStrategy,
     markers: &'a MarkerEnvironment,
     python_requirement: PythonRequirement,
     selector: CandidateSelector,
     index: &'a InMemoryIndex,
     installed_packages: &'a InstalledPackages,
-    /// Incompatibilities for packages that are entirely unavailable
+    /// Incompatibilities for packages that are entirely unavailable.
     unavailable_packages: DashMap<PackageName, UnavailablePackage>,
+    /// Incompatibilities for packages that are unavailable at specific versions.
+    incomplete_packages: DashMap<PackageName, DashMap<Version, IncompletePackage>>,
     /// The set of all registry-based packages visited during resolution.
     visited: DashSet<PackageName>,
     reporter: Option<Arc<dyn Reporter>>,
@@ -140,6 +156,7 @@ impl<
         client: &'a RegistryClient,
         flat_index: &'a FlatIndex,
         index: &'a InMemoryIndex,
+        hasher: &'a HashStrategy,
         build_context: &'a Context,
         installed_packages: &'a InstalledPackages,
     ) -> Result<Self, ResolveError> {
@@ -150,6 +167,7 @@ impl<
             tags,
             PythonRequirement::new(interpreter, markers),
             AllowedYanks::from_manifest(&manifest, markers),
+            hasher,
             options.exclude_newer,
             build_context.no_binary(),
             build_context.no_build(),
@@ -157,6 +175,7 @@ impl<
         Self::new_custom_io(
             manifest,
             options,
+            hasher,
             markers,
             PythonRequirement::new(interpreter, markers),
             index,
@@ -173,9 +192,11 @@ impl<
     > Resolver<'a, Provider, InstalledPackages>
 {
     /// Initialize a new resolver using a user provided backend.
+    #[allow(clippy::too_many_arguments)]
     pub fn new_custom_io(
         manifest: Manifest,
         options: Options,
+        hasher: &'a HashStrategy,
         markers: &'a MarkerEnvironment,
         python_requirement: PythonRequirement,
         index: &'a InMemoryIndex,
@@ -185,6 +206,7 @@ impl<
         Ok(Self {
             index,
             unavailable_packages: DashMap::default(),
+            incomplete_packages: DashMap::default(),
             visited: DashSet::default(),
             selector: CandidateSelector::for_resolution(options, &manifest, markers),
             dependency_mode: options.dependency_mode,
@@ -197,6 +219,7 @@ impl<
             preferences: Preferences::from_iter(manifest.preferences, markers),
             exclusions: manifest.exclusions,
             editables: Editables::from_requirements(manifest.editables),
+            hasher,
             markers,
             python_requirement,
             reporter: None,
@@ -247,7 +270,8 @@ impl<
                         .with_selector(self.selector.clone())
                         .with_python_requirement(&self.python_requirement)
                         .with_index_locations(self.provider.index_locations())
-                        .with_unavailable_packages(&self.unavailable_packages),
+                        .with_unavailable_packages(&self.unavailable_packages)
+                        .with_incomplete_packages(&self.incomplete_packages),
                     )
                 } else {
                     err
@@ -292,13 +316,13 @@ impl<
             }
 
             // Choose a package version.
-            let Some(highest_priority_pkg) =
-                state
-                    .partial_solution
-                    .pick_highest_priority_pkg(|package, _range| {
-                        priorities.get(package).unwrap_or_default()
-                    })
+            let Some(highest_priority_pkg) = state
+                .partial_solution
+                .pick_highest_priority_pkg(|package, _range| priorities.get(package))
             else {
+                if enabled!(Level::DEBUG) {
+                    prefetcher.log_tried_versions();
+                }
                 let selection = state.partial_solution.extract_solution();
                 return ResolutionGraph::from_state(
                     &selection,
@@ -352,10 +376,10 @@ impl<
                                     UnavailablePackage::NotFound => {
                                         "was not found in the package registry"
                                     }
-                                    UnavailablePackage::InvalidMetadata => {
+                                    UnavailablePackage::InvalidMetadata(_) => {
                                         "was found, but the metadata could not be parsed"
                                     }
-                                    UnavailablePackage::InvalidStructure => {
+                                    UnavailablePackage::InvalidStructure(_) => {
                                         "was found, but has an invalid format"
                                     }
                                 })
@@ -458,6 +482,9 @@ impl<
                             .iter()
                             .any(|(dependency, _)| dependency == package) =>
                     {
+                        if enabled!(Level::DEBUG) {
+                            prefetcher.log_tried_versions();
+                        }
                         return Err(PubGrubError::SelfDependency {
                             package: package.clone(),
                             version: version.clone(),
@@ -493,26 +520,31 @@ impl<
     async fn visit_package(
         &self,
         package: &PubGrubPackage,
-        priorities: &mut PubGrubPriorities,
         request_sink: &tokio::sync::mpsc::Sender<Request>,
     ) -> Result<(), ResolveError> {
         match package {
             PubGrubPackage::Root(_) => {}
             PubGrubPackage::Python(_) => {}
-            PubGrubPackage::Package(package_name, _extra, None) => {
+            PubGrubPackage::Package(name, _extra, None) => {
+                // Verify that the package is allowed under the hash-checking policy.
+                if !self.hasher.allows_package(name) {
+                    return Err(ResolveError::UnhashedPackage(name.clone()));
+                }
+
                 // Emit a request to fetch the metadata for this package.
-                if self.index.packages.register(package_name.clone()) {
-                    priorities.add(package_name.clone());
-                    request_sink
-                        .send(Request::Package(package_name.clone()))
-                        .await?;
+                if self.index.packages.register(name.clone()) {
+                    request_sink.send(Request::Package(name.clone())).await?;
                 }
             }
-            PubGrubPackage::Package(package_name, _extra, Some(url)) => {
+            PubGrubPackage::Package(name, _extra, Some(url)) => {
+                // Verify that the package is allowed under the hash-checking policy.
+                if !self.hasher.allows_url(url) {
+                    return Err(ResolveError::UnhashedPackage(name.clone()));
+                }
+
                 // Emit a request to fetch the metadata for this distribution.
-                let dist = Dist::from_url(package_name.clone(), url.clone())?;
-                if self.index.distributions.register(dist.package_id()) {
-                    priorities.add(dist.name().clone());
+                let dist = Dist::from_url(name.clone(), url.clone())?;
+                if self.index.distributions.register(dist.version_id()) {
                     request_sink.send(Request::Dist(dist)).await?;
                 }
             }
@@ -612,26 +644,37 @@ impl<
                 let response = self
                     .index
                     .distributions
-                    .wait(&dist.package_id())
+                    .wait(&dist.version_id())
                     .await
                     .ok_or(ResolveError::Unregistered)?;
 
                 // If we failed to fetch the metadata for a URL, we can't proceed.
                 let metadata = match &*response {
-                    MetadataResponse::Found(metadata) => metadata,
+                    MetadataResponse::Found(archive) => &archive.metadata,
                     MetadataResponse::Offline => {
                         self.unavailable_packages
                             .insert(package_name.clone(), UnavailablePackage::Offline);
                         return Ok(None);
                     }
-                    MetadataResponse::InvalidMetadata(_) => {
-                        self.unavailable_packages
-                            .insert(package_name.clone(), UnavailablePackage::InvalidMetadata);
+                    MetadataResponse::InvalidMetadata(err) => {
+                        self.unavailable_packages.insert(
+                            package_name.clone(),
+                            UnavailablePackage::InvalidMetadata(err.to_string()),
+                        );
                         return Ok(None);
                     }
-                    MetadataResponse::InvalidStructure(_) => {
-                        self.unavailable_packages
-                            .insert(package_name.clone(), UnavailablePackage::InvalidStructure);
+                    MetadataResponse::InconsistentMetadata(err) => {
+                        self.unavailable_packages.insert(
+                            package_name.clone(),
+                            UnavailablePackage::InvalidMetadata(err.to_string()),
+                        );
+                        return Ok(None);
+                    }
+                    MetadataResponse::InvalidStructure(err) => {
+                        self.unavailable_packages.insert(
+                            package_name.clone(),
+                            UnavailablePackage::InvalidStructure(err.to_string()),
+                        );
                         return Ok(None);
                     }
                 };
@@ -751,7 +794,7 @@ impl<
                 let version = candidate.version().clone();
 
                 // Emit a request to fetch the metadata for this version.
-                if self.index.distributions.register(candidate.package_id()) {
+                if self.index.distributions.register(candidate.version_id()) {
                     let request = match dist.for_resolution() {
                         ResolvedDistRef::Installable(dist) => Request::Dist(dist.clone()),
                         ResolvedDistRef::Installed(dist) => Request::Installed(dist.clone()),
@@ -796,9 +839,11 @@ impl<
                 for (package, version) in constraints.iter() {
                     debug!("Adding direct dependency: {package}{version}");
 
+                    // Update the package priorities.
+                    priorities.insert(package, version);
+
                     // Emit a request to fetch the metadata for this package.
-                    self.visit_package(package, priorities, request_sink)
-                        .await?;
+                    self.visit_package(package, request_sink).await?;
                 }
 
                 // Add a dependency on each editable.
@@ -835,13 +880,13 @@ impl<
                             Some(url) => PubGrubDistribution::from_url(package_name, url),
                             None => PubGrubDistribution::from_registry(package_name, version),
                         };
-                        let package_id = dist.package_id();
+                        let version_id = dist.version_id();
 
                         // Wait for the metadata to be available.
                         self.index
                             .distributions
-                            .wait(&package_id)
-                            .instrument(info_span!("distributions_wait", %package_id))
+                            .wait(&version_id)
+                            .instrument(info_span!("distributions_wait", %version_id))
                             .await
                             .ok_or(ResolveError::Unregistered)?;
                     }
@@ -865,9 +910,11 @@ impl<
                     for (dep_package, dep_version) in constraints.iter() {
                         debug!("Adding transitive dependency for {package}{version}: {dep_package}{dep_version}");
 
+                        // Update the package priorities.
+                        priorities.insert(dep_package, dep_version);
+
                         // Emit a request to fetch the metadata for this package.
-                        self.visit_package(dep_package, priorities, request_sink)
-                            .await?;
+                        self.visit_package(dep_package, request_sink).await?;
                     }
 
                     // If a package has an extra, insert a constraint on the base package.
@@ -886,7 +933,7 @@ impl<
                     Some(url) => PubGrubDistribution::from_url(package_name, url),
                     None => PubGrubDistribution::from_registry(package_name, version),
                 };
-                let package_id = dist.package_id();
+                let version_id = dist.version_id();
 
                 // If the package does not exist in the registry or locally, we cannot fetch its dependencies
                 if self.unavailable_packages.get(package_name).is_some()
@@ -908,25 +955,58 @@ impl<
                 let response = self
                     .index
                     .distributions
-                    .wait(&package_id)
-                    .instrument(info_span!("distributions_wait", %package_id))
+                    .wait(&version_id)
+                    .instrument(info_span!("distributions_wait", %version_id))
                     .await
                     .ok_or(ResolveError::Unregistered)?;
 
-                let metadata = match *response {
-                    MetadataResponse::Found(ref metadata) => metadata,
+                let metadata = match &*response {
+                    MetadataResponse::Found(archive) => &archive.metadata,
                     MetadataResponse::Offline => {
+                        self.incomplete_packages
+                            .entry(package_name.clone())
+                            .or_default()
+                            .insert(version.clone(), IncompletePackage::Offline);
                         return Ok(Dependencies::Unavailable(
                             "network connectivity is disabled, but the metadata wasn't found in the cache"
                                 .to_string(),
                         ));
                     }
-                    MetadataResponse::InvalidMetadata(_) => {
+                    MetadataResponse::InvalidMetadata(err) => {
+                        warn!("Unable to extract metadata for {package_name}: {err}");
+                        self.incomplete_packages
+                            .entry(package_name.clone())
+                            .or_default()
+                            .insert(
+                                version.clone(),
+                                IncompletePackage::InvalidMetadata(err.to_string()),
+                            );
                         return Ok(Dependencies::Unavailable(
                             "the package metadata could not be parsed".to_string(),
                         ));
                     }
-                    MetadataResponse::InvalidStructure(_) => {
+                    MetadataResponse::InconsistentMetadata(err) => {
+                        warn!("Unable to extract metadata for {package_name}: {err}");
+                        self.incomplete_packages
+                            .entry(package_name.clone())
+                            .or_default()
+                            .insert(
+                                version.clone(),
+                                IncompletePackage::InconsistentMetadata(err.to_string()),
+                            );
+                        return Ok(Dependencies::Unavailable(
+                            "the package metadata was inconsistent".to_string(),
+                        ));
+                    }
+                    MetadataResponse::InvalidStructure(err) => {
+                        warn!("Unable to extract metadata for {package_name}: {err}");
+                        self.incomplete_packages
+                            .entry(package_name.clone())
+                            .or_default()
+                            .insert(
+                                version.clone(),
+                                IncompletePackage::InvalidStructure(err.to_string()),
+                            );
                         return Ok(Dependencies::Unavailable(
                             "the package has an invalid format".to_string(),
                         ));
@@ -947,9 +1027,11 @@ impl<
                 for (package, version) in constraints.iter() {
                     debug!("Adding transitive dependency: {package}{version}");
 
+                    // Update the package priorities.
+                    priorities.insert(package, version);
+
                     // Emit a request to fetch the metadata for this package.
-                    self.visit_package(package, priorities, request_sink)
-                        .await?;
+                    self.visit_package(package, request_sink).await?;
                 }
 
                 // If a package has an extra, insert a constraint on the base package.
@@ -982,9 +1064,10 @@ impl<
                 }
                 Some(Response::Installed { dist, metadata }) => {
                     trace!("Received installed distribution metadata for: {dist}");
-                    self.index
-                        .distributions
-                        .done(dist.package_id(), MetadataResponse::Found(metadata));
+                    self.index.distributions.done(
+                        dist.version_id(),
+                        MetadataResponse::Found(ArchiveMetadata::from(metadata)),
+                    );
                 }
                 Some(Response::Dist {
                     dist: Dist::Built(dist),
@@ -1000,7 +1083,7 @@ impl<
                         }
                         _ => {}
                     }
-                    self.index.distributions.done(dist.package_id(), metadata);
+                    self.index.distributions.done(dist.version_id(), metadata);
                 }
                 Some(Response::Dist {
                     dist: Dist::Source(dist),
@@ -1016,7 +1099,7 @@ impl<
                         }
                         _ => {}
                     }
-                    self.index.distributions.done(dist.package_id(), metadata);
+                    self.index.distributions.done(dist.version_id(), metadata);
                 }
                 None => {}
             }
@@ -1121,7 +1204,7 @@ impl<
                 };
 
                 // Emit a request to fetch the metadata for this version.
-                if self.index.distributions.register(candidate.package_id()) {
+                if self.index.distributions.register(candidate.version_id()) {
                     let dist = dist.for_resolution().to_owned();
 
                     let response = match dist {
