@@ -1,9 +1,10 @@
 use std::fmt::Write;
+
 use std::path::Path;
 
 use anstream::eprint;
 use anyhow::{anyhow, Context, Result};
-use chrono::{DateTime, Utc};
+
 use itertools::Itertools;
 use owo_colors::OwoColorize;
 use tempfile::tempdir_in;
@@ -18,12 +19,11 @@ use pep508_rs::{MarkerEnvironment, Requirement};
 use platform_tags::Tags;
 use pypi_types::{Metadata23, Yanked};
 use requirements_txt::EditableRequirement;
-use uv_auth::{KeyringProvider, GLOBAL_AUTH_STORE};
 use uv_cache::Cache;
 use uv_client::{
-    BaseClientBuilder, Connectivity, FlatIndex, FlatIndexClient, RegistryClient,
-    RegistryClientBuilder,
+    BaseClientBuilder, Connectivity, FlatIndexClient, RegistryClient, RegistryClientBuilder,
 };
+use uv_configuration::KeyringProviderType;
 use uv_configuration::{
     ConfigSettings, Constraints, IndexStrategy, NoBinary, NoBuild, Overrides, Reinstall,
     SetupPyStrategy, Upgrade,
@@ -38,10 +38,10 @@ use uv_requirements::{
     RequirementsSpecification, SourceTreeResolver,
 };
 use uv_resolver::{
-    DependencyMode, Exclusions, InMemoryIndex, Manifest, Options, OptionsBuilder, PreReleaseMode,
-    Preference, ResolutionGraph, ResolutionMode, Resolver,
+    DependencyMode, ExcludeNewer, Exclusions, FlatIndex, InMemoryIndex, Manifest, Options,
+    OptionsBuilder, PreReleaseMode, Preference, ResolutionGraph, ResolutionMode, Resolver,
 };
-use uv_types::{BuildIsolation, InFlight};
+use uv_types::{BuildIsolation, HashStrategy, InFlight};
 use uv_warnings::warn_user;
 
 use crate::commands::reporters::{DownloadReporter, InstallReporter, ResolverReporter};
@@ -63,10 +63,11 @@ pub(crate) async fn pip_install(
     upgrade: Upgrade,
     index_locations: IndexLocations,
     index_strategy: IndexStrategy,
-    keyring_provider: KeyringProvider,
+    keyring_provider: KeyringProviderType,
     reinstall: Reinstall,
     link_mode: LinkMode,
     compile: bool,
+    require_hashes: bool,
     setup_py: SetupPyStrategy,
     connectivity: Connectivity,
     config_settings: &ConfigSettings,
@@ -74,7 +75,7 @@ pub(crate) async fn pip_install(
     no_build: NoBuild,
     no_binary: NoBinary,
     strict: bool,
-    exclude_newer: Option<DateTime<Utc>>,
+    exclude_newer: Option<ExcludeNewer>,
     python: Option<String>,
     system: bool,
     break_system_packages: bool,
@@ -84,10 +85,11 @@ pub(crate) async fn pip_install(
     printer: Printer,
 ) -> Result<ExitStatus> {
     let start = std::time::Instant::now();
+
     let client_builder = BaseClientBuilder::new()
         .connectivity(connectivity)
         .native_tls(native_tls)
-        .keyring_provider(keyring_provider);
+        .keyring(keyring_provider);
 
     // Read all requirements from the provided sources.
     let RequirementsSpecification {
@@ -158,6 +160,7 @@ pub(crate) async fn pip_install(
     if reinstall.is_none()
         && upgrade.is_none()
         && source_trees.is_empty()
+        && overrides.is_empty()
         && site_packages.satisfies(&requirements, &editables, &constraints)?
     {
         let num_requirements = requirements.len() + editables.len();
@@ -183,22 +186,30 @@ pub(crate) async fn pip_install(
     let tags = venv.interpreter().tags()?;
     let markers = venv.interpreter().markers();
 
+    // Collect the set of required hashes.
+    let hasher = if require_hashes {
+        HashStrategy::from_requirements(
+            requirements
+                .iter()
+                .chain(overrides.iter())
+                .map(|entry| (&entry.requirement, entry.hashes.as_slice())),
+            markers,
+        )?
+    } else {
+        HashStrategy::None
+    };
+
     // Incorporate any index locations from the provided sources.
     let index_locations =
         index_locations.combine(index_url, extra_index_urls, find_links, no_index);
 
-    // Add all authenticated sources to the store.
-    for url in index_locations.urls() {
-        GLOBAL_AUTH_STORE.save_from_url(url);
-    }
-
     // Initialize the registry client.
     let client = RegistryClientBuilder::new(cache.clone())
         .native_tls(native_tls)
         .connectivity(connectivity)
         .index_urls(index_locations.index_urls())
         .index_strategy(index_strategy)
-        .keyring_provider(keyring_provider)
+        .keyring(keyring_provider)
         .markers(markers)
         .platform(interpreter.platform())
         .build();
@@ -207,7 +218,7 @@ pub(crate) async fn pip_install(
     let flat_index = {
         let client = FlatIndexClient::new(&client, &cache);
         let entries = client.fetch(index_locations.flat_index()).await?;
-        FlatIndex::from_entries(entries, tags, &no_build, &no_binary)
+        FlatIndex::from_entries(entries, tags, &hasher, &no_build, &no_binary)
     };
 
     // Determine whether to enable build isolation.
@@ -239,6 +250,7 @@ pub(crate) async fn pip_install(
         setup_py,
         config_settings,
         build_isolation,
+        link_mode,
         &no_build,
         &no_binary,
     )
@@ -247,25 +259,44 @@ pub(crate) async fn pip_install(
     // Resolve the requirements from the provided sources.
     let requirements = {
         // Convert from unnamed to named requirements.
-        let mut requirements =
-            NamedRequirementsResolver::new(requirements, &resolve_dispatch, &client, &index)
-                .with_reporter(ResolverReporter::from(printer))
-                .resolve()
-                .await?;
+        let mut requirements = NamedRequirementsResolver::new(
+            requirements,
+            &hasher,
+            &resolve_dispatch,
+            &client,
+            &index,
+        )
+        .with_reporter(ResolverReporter::from(printer))
+        .resolve()
+        .await?;
 
         // Resolve any source trees into requirements.
         if !source_trees.is_empty() {
             requirements.extend(
-                SourceTreeResolver::new(source_trees, extras, &resolve_dispatch, &client, &index)
-                    .with_reporter(ResolverReporter::from(printer))
-                    .resolve()
-                    .await?,
+                SourceTreeResolver::new(
+                    source_trees,
+                    extras,
+                    &hasher,
+                    &resolve_dispatch,
+                    &client,
+                    &index,
+                )
+                .with_reporter(ResolverReporter::from(printer))
+                .resolve()
+                .await?,
             );
         }
 
         requirements
     };
 
+    // Resolve the overrides from the provided sources.
+    let overrides =
+        NamedRequirementsResolver::new(overrides, &hasher, &resolve_dispatch, &client, &index)
+            .with_reporter(ResolverReporter::from(printer))
+            .resolve()
+            .await?;
+
     // Build all editable distributions. The editables are shared between resolution and
     // installation, and should live for the duration of the command. If an editable is already
     // installed in the environment, we'll still re-build it here.
@@ -277,6 +308,7 @@ pub(crate) async fn pip_install(
         build_editables(
             &editables,
             editable_wheel_dir.path(),
+            &hasher,
             &cache,
             &interpreter,
             tags,
@@ -301,6 +333,7 @@ pub(crate) async fn pip_install(
         overrides,
         project,
         &editables,
+        &hasher,
         &site_packages,
         &reinstall,
         &upgrade,
@@ -345,6 +378,7 @@ pub(crate) async fn pip_install(
             setup_py,
             config_settings,
             build_isolation,
+            link_mode,
             &no_build,
             &no_binary,
         )
@@ -361,6 +395,7 @@ pub(crate) async fn pip_install(
         link_mode,
         compile,
         &index_locations,
+        &hasher,
         tags,
         &client,
         &in_flight,
@@ -436,6 +471,7 @@ async fn read_requirements(
 async fn build_editables(
     editables: &[EditableRequirement],
     editable_wheel_dir: &Path,
+    hasher: &HashStrategy,
     cache: &Cache,
     interpreter: &Interpreter,
     tags: &Tags,
@@ -445,7 +481,7 @@ async fn build_editables(
 ) -> Result<Vec<BuiltEditable>, Error> {
     let start = std::time::Instant::now();
 
-    let downloader = Downloader::new(cache, tags, client, build_dispatch)
+    let downloader = Downloader::new(cache, tags, hasher, client, build_dispatch)
         .with_reporter(DownloadReporter::from(printer).with_length(editables.len() as u64));
 
     let editables = LocalEditables::from_editables(editables.iter().map(|editable| {
@@ -502,6 +538,7 @@ async fn resolve(
     overrides: Vec<Requirement>,
     project: Option<PackageName>,
     editables: &[BuiltEditable],
+    hasher: &HashStrategy,
     site_packages: &SitePackages<'_>,
     reinstall: &Reinstall,
     upgrade: &Upgrade,
@@ -548,6 +585,7 @@ async fn resolve(
         &constraints,
         &overrides,
         &editables,
+        hasher,
         build_dispatch,
         client,
         index,
@@ -578,6 +616,7 @@ async fn resolve(
         client,
         flat_index,
         index,
+        hasher,
         build_dispatch,
         site_packages,
     )?
@@ -621,6 +660,7 @@ async fn install(
     link_mode: LinkMode,
     compile: bool,
     index_urls: &IndexLocations,
+    hasher: &HashStrategy,
     tags: &Tags,
     client: &RegistryClient,
     in_flight: &InFlight,
@@ -648,6 +688,7 @@ async fn install(
             site_packages,
             reinstall,
             no_binary,
+            hasher,
             index_urls,
             cache,
             venv,
@@ -700,7 +741,7 @@ async fn install(
     } else {
         let start = std::time::Instant::now();
 
-        let downloader = Downloader::new(cache, tags, client, build_dispatch)
+        let downloader = Downloader::new(cache, tags, hasher, client, build_dispatch)
             .with_reporter(DownloadReporter::from(printer).with_length(remote.len() as u64));
 
         let wheels = downloader
@@ -1016,6 +1057,9 @@ enum Error {
     #[error(transparent)]
     Platform(#[from] platform_tags::PlatformError),
 
+    #[error(transparent)]
+    Hash(#[from] uv_types::HashStrategyError),
+
     #[error(transparent)]
     Io(#[from] std::io::Error),
 
