@@ -10,15 +10,15 @@ use distribution_types::{
     IndexLocations, InstalledMetadata, LocalDist, LocalEditable, LocalEditables, Name, ResolvedDist,
 };
 use install_wheel_rs::linker::LinkMode;
+
 use platform_tags::Tags;
 use pypi_types::Yanked;
 use requirements_txt::EditableRequirement;
-use uv_auth::{KeyringProvider, GLOBAL_AUTH_STORE};
 use uv_cache::{ArchiveTarget, ArchiveTimestamp, Cache};
 use uv_client::{
-    BaseClientBuilder, Connectivity, FlatIndex, FlatIndexClient, RegistryClient,
-    RegistryClientBuilder,
+    BaseClientBuilder, Connectivity, FlatIndexClient, RegistryClient, RegistryClientBuilder,
 };
+use uv_configuration::KeyringProviderType;
 use uv_configuration::{
     ConfigSettings, IndexStrategy, NoBinary, NoBuild, Reinstall, SetupPyStrategy,
 };
@@ -30,8 +30,8 @@ use uv_requirements::{
     ExtrasSpecification, NamedRequirementsResolver, RequirementsSource, RequirementsSpecification,
     SourceTreeResolver,
 };
-use uv_resolver::{DependencyMode, InMemoryIndex, Manifest, OptionsBuilder, Resolver};
-use uv_types::{BuildIsolation, EmptyInstalledPackages, InFlight};
+use uv_resolver::{DependencyMode, FlatIndex, InMemoryIndex, Manifest, OptionsBuilder, Resolver};
+use uv_types::{BuildIsolation, EmptyInstalledPackages, HashStrategy, InFlight};
 use uv_warnings::warn_user;
 
 use crate::commands::reporters::{DownloadReporter, InstallReporter, ResolverReporter};
@@ -45,9 +45,10 @@ pub(crate) async fn pip_sync(
     reinstall: &Reinstall,
     link_mode: LinkMode,
     compile: bool,
+    require_hashes: bool,
     index_locations: IndexLocations,
     index_strategy: IndexStrategy,
-    keyring_provider: KeyringProvider,
+    keyring_provider: KeyringProviderType,
     setup_py: SetupPyStrategy,
     connectivity: Connectivity,
     config_settings: &ConfigSettings,
@@ -67,7 +68,7 @@ pub(crate) async fn pip_sync(
     let client_builder = BaseClientBuilder::new()
         .connectivity(connectivity)
         .native_tls(native_tls)
-        .keyring_provider(keyring_provider);
+        .keyring(keyring_provider);
 
     // Read all requirements from the provided sources.
     let RequirementsSpecification {
@@ -131,23 +132,31 @@ pub(crate) async fn pip_sync(
 
     // Determine the current environment markers.
     let tags = venv.interpreter().tags()?;
+    let markers = venv.interpreter().markers();
+
+    // Collect the set of required hashes.
+    let hasher = if require_hashes {
+        HashStrategy::from_requirements(
+            requirements
+                .iter()
+                .map(|entry| (&entry.requirement, entry.hashes.as_slice())),
+            markers,
+        )?
+    } else {
+        HashStrategy::None
+    };
 
     // Incorporate any index locations from the provided sources.
     let index_locations =
         index_locations.combine(index_url, extra_index_urls, find_links, no_index);
 
-    // Add all authenticated sources to the store.
-    for url in index_locations.urls() {
-        GLOBAL_AUTH_STORE.save_from_url(url);
-    }
-
     // Initialize the registry client.
     let client = RegistryClientBuilder::new(cache.clone())
         .native_tls(native_tls)
         .connectivity(connectivity)
         .index_urls(index_locations.index_urls())
         .index_strategy(index_strategy)
-        .keyring_provider(keyring_provider)
+        .keyring(keyring_provider)
         .markers(venv.interpreter().markers())
         .platform(venv.interpreter().platform())
         .build();
@@ -156,7 +165,7 @@ pub(crate) async fn pip_sync(
     let flat_index = {
         let client = FlatIndexClient::new(&client, &cache);
         let entries = client.fetch(index_locations.flat_index()).await?;
-        FlatIndex::from_entries(entries, tags, &no_build, &no_binary)
+        FlatIndex::from_entries(entries, tags, &hasher, &no_build, &no_binary)
     };
 
     // Create a shared in-memory index.
@@ -191,6 +200,7 @@ pub(crate) async fn pip_sync(
         setup_py,
         config_settings,
         build_isolation,
+        link_mode,
         &no_build,
         &no_binary,
     );
@@ -199,7 +209,7 @@ pub(crate) async fn pip_sync(
     let requirements = {
         // Convert from unnamed to named requirements.
         let mut requirements =
-            NamedRequirementsResolver::new(requirements, &build_dispatch, &client, &index)
+            NamedRequirementsResolver::new(requirements, &hasher, &build_dispatch, &client, &index)
                 .with_reporter(ResolverReporter::from(printer))
                 .resolve()
                 .await?;
@@ -210,6 +220,7 @@ pub(crate) async fn pip_sync(
                 SourceTreeResolver::new(
                     source_trees,
                     &ExtrasSpecification::None,
+                    &hasher,
                     &build_dispatch,
                     &client,
                     &index,
@@ -228,6 +239,7 @@ pub(crate) async fn pip_sync(
         editables,
         &site_packages,
         reinstall,
+        &hasher,
         venv.interpreter(),
         tags,
         &cache,
@@ -251,6 +263,7 @@ pub(crate) async fn pip_sync(
             site_packages,
             reinstall,
             &no_binary,
+            &hasher,
             &index_locations,
             &cache,
             &venv,
@@ -304,6 +317,7 @@ pub(crate) async fn pip_sync(
             &client,
             &flat_index,
             &index,
+            &hasher,
             &build_dispatch,
             // TODO(zanieb): We should consider support for installed packages in pip sync
             &EmptyInstalledPackages,
@@ -347,7 +361,7 @@ pub(crate) async fn pip_sync(
     } else {
         let start = std::time::Instant::now();
 
-        let downloader = Downloader::new(&cache, tags, &client, &build_dispatch)
+        let downloader = Downloader::new(&cache, tags, &hasher, &client, &build_dispatch)
             .with_reporter(DownloadReporter::from(printer).with_length(remote.len() as u64));
 
         let wheels = downloader
@@ -543,6 +557,7 @@ async fn resolve_editables(
     editables: Vec<EditableRequirement>,
     site_packages: &SitePackages<'_>,
     reinstall: &Reinstall,
+    hasher: &HashStrategy,
     interpreter: &Interpreter,
     tags: &Tags,
     cache: &Cache,
@@ -609,7 +624,7 @@ async fn resolve_editables(
     } else {
         let start = std::time::Instant::now();
 
-        let downloader = Downloader::new(cache, tags, client, build_dispatch)
+        let downloader = Downloader::new(cache, tags, hasher, client, build_dispatch)
             .with_reporter(DownloadReporter::from(printer).with_length(uninstalled.len() as u64));
 
         let editables = LocalEditables::from_editables(uninstalled.iter().map(|editable| {
